# File skyline_3.py
from backtrack import Backtrack
import sys

class Skyline(Backtrack):
 
    vector = []    
    solutionNr = 0

    # Siehe https://puzzlephil.com/puzzles/skyline/de/
   

    def __init__(self, high:int=10, skyline:dict={}, test:bool = False, showAttempts:bool = False):        
        
        self.high = high
        self.skyline = skyline
        self.test = test
        self.showAttempts = showAttempts

        if not(len(self.skyline["west"]) == len(self.skyline["east"]) == len(self.skyline["north"]) == len(self.skyline["south"])):
               print("Ungleiche Randspalten - Programm beendet")
               sys.exit()
        self.size = len(self.skyline["west"])
        self.vector = [0 for i in range(self.size*self.size)]
 
        self.rset = [set() for i in range(self.size)]
        self.cset = [set() for i in range(self.size)]
        self.values = set(range(self.high, (self.size+1)*self.high, self.high))
        super().__init__(test, showAttempts)

    def rowAndCol(self, where:int):
        s = self.size
        r = where // s
        c = where % s
        return (r, c)

    def moves(self, level:int) -> list[int]:
        r, c = self.rowAndCol(level)
        return list(self.values - self.rset[r] - self.cset[c])     

    def valid(self, level:int, move:int) -> bool:
        r, c = self.rowAndCol(level)
        s = self.size

        # Jede der vier Richtungen muss einzeln geprüft werden
        # Daher ist der Einsatz einer eigenen Funktion check sinnvoll
        # Argumente:
        # visible: wie viele Hochhäuser sind sichtbar?
        # sl: Argumente für slice
        def check(visible:int, sl:list[int]) -> bool:
            
            if visible==0:
                # Unbekannte Anzahl, daher ist die Annahme nicht falsch
                return True
            
            max = 0
            count = 0

            # Wenn sl[2] gleich 1 oder -1 ist, wird eine Zeile untersucht
            rowtest = abs(sl[2])==1   
            rc = c if rowtest else r

            # Letzte Spalte oder letzte Zeile erreicht?
            if rc+1 == s:
                # Zählen der sichtbaren Häuser
                for v in self.vector[slice(*sl)]:
                    if v > max:
                        # Ist das Haus höher als die vorhergehenden?
                        # Dann count um 1 erhöhen
                        max = v
                        count += 1

                # Sollten Zwischenergebnisse zum Testen ausgegeben werden?
                if self.test:
                    self.show(f"level={level} r={r} c={c} move={move} attempt={self.attempts}"
                              f"\nrowtest={rowtest} max={max} count={count} visible={visible} \n"
                              f"slice={sl} ==> {self.vector[slice(*sl)]}")

                # Wurden exakt die Häuser laut Angabe gefunden?
                if count != visible:
                    return False                       
                                            
            return True
        
        # Probeweise das neue Haus einfügen
        self.vector[r*s + c] = move  

        # Probeweise das neue Haus einfügen
        res = all([
            check(self.skyline["west"][r],  [r*s, (r+1)*s, 1]),
            check(self.skyline["east"][r],  [-(s-r-1)*s-1, -(s-r)*s-1, -1]),
            check(self.skyline["north"][c], [c, c+s*s, s]),
            check(self.skyline["south"][c], [-(s-c), -((s+1)*s-c), -s])
        ])
        
        if self.test and (r+1==self.size or c+1==self.size) and res:
            self.show(f"valid={res}, level={level}, r={r}, c={c},  move={move}, attempt={self.attempts}")

        # Das probeweise gesetzte Haus wieder entfernen
        # Falls es gepasst hat, wird es von record gespeichert
        self.vector[r*s + c] = 0  

        return res
            
    def record(self, level:int, move:int):
        self.vector[level]=move
        r, c = self.rowAndCol(level)
        self.rset[r].add(move)  # Wert zur Menge der Zeilenwerte hinzufügen
        self.cset[c].add(move)  # Wert zur Menge der Spaltenwerte hinzufügen
        

    def undo(self, level:int, move:int):
        self.vector[level]=0     
        r, c = self.rowAndCol(level)
        self.rset[r].remove(move)  # Wert aus der Menge der Zeilenwerte entfernen
        self.cset[c].remove(move)  # Wert aus der Menge der Spaltenwerte entfernen        

    def done(self, level: int)->bool:
        # Wurde das letzte Haus bestimmt?  
        if level==self.size*self.size-1:
            self.solutionNr += 1
            self.show(f"Lösung {self.solutionNr:4d}, {self.attempts:18_d}. Versuch")
            return True
        return False

    def show(self, text:str):
        print(f"\n{text}")
        for r in range(self.size):
            for c in range(self.size):
                print(f"{self.vector[r*self.size + c]:4d}", end="")
            print()

if __name__ == "__main__":

         
    skyline = {
        "west": [1,3,3,2],
        "east": [2,2,1,3],
        "north":[1,3,2,2],
        "south":[2,1,3,2]
    }
    Skyline(skyline=skyline, test=False, showAttempts=True)    
