# File skyline_4.py
from backtrack import Backtrack
import sys

class Skyline(Backtrack):
 
    vector = []    
    resultatnumber = 0

    def __init__(self, high=10, skyline={}, known=[], test:bool = False, showAttempts:bool = False):        
        
        self.high = high
        self.skyline = skyline
        self.known = known
        self.test = test
        self.showAttempts = showAttempts

        if not(len(self.skyline["west"]) == len(self.skyline["east"]) == len(self.skyline["north"]) == len(self.skyline["south"])):
               print("Ungleiche Randspalten - Programm beendet")
               sys.exit()
               
        self.size = len(self.skyline["west"])
        self.vector = [0 for i in range(self.size*self.size)]
        self.rset = [set() for i in range(self.size)]
        self.cset = [set() for i in range(self.size)]

        # Vorgegebene Werte eintragen:
        for k in self.known:
            r, c = self.rowAndCol(k)
            self.rset[r].add(self.known[k])
            self.cset[c].add(self.known[k])
            
        self.values = set(range(high, (self.size+1)*high, high))
        super().__init__(test, showAttempts)

    def rowAndCol(self, where:int):
        s = self.size
        r = where // s
        c = where % s
        return (r, c)

    def moves(self, level:int):
        if v:=self.known.get(level): # Vorgegebenes Feld
           return [v]
        r, c = self.rowAndCol(level)
        return list(self.values - self.rset[r] - self.cset[c])       

    def valid(self, level:int, move:int) -> bool:
        r, c = self.rowAndCol(level)
        s = self.size
        
        # Jede der vier Richtungen muss einzeln geprüft werden
        # Daher ist der Einsatz einer eigenen Funktion check sinnvoll
        # Argumente:
        # visible: wie viele Hochhäuser sind sichtbar?
        # sl: Argumente für slice
        def check(visible:int, sl):
            
            if visible==0:
                # Unbekannte Anzahl, daher ist die Annahme nicht falsch
                return True
            
            max = 0
            count = 0
           
            # Zählen der sichtbaren Häuser
            for v in self.vector[slice(*sl)]:
                if v > max:
                    max = v
                    count += 1 
            
            # Ist das Ende der Zeile oder Spalte erreicht,
            # muss count mit dem vorgegeben Wert visible exakt übereinstimmen
            # Zur Testausgabe werden die zwei möglichen Fälle als case gespeichert
            case = 0

            # Zeile oder Spalte untersuchen?
            rowtest = abs(sl[2])==1 # Schrittweite == 1 ==> Zeile  
           
            if rowtest:   
                if c+1==s and count != visible:  # letzte Spalte erreicht?
                    case = 1                
            else:
                if r+1==s and count != visible:  # letzte Zeile erreicht?
                    case = 2               
                    
            # Sollten Zwischenergebnisse zum Testen ausgegeben werden?
            if self.test:
                self.show(f"level={level} r={r} c={c} move={move} attempt={self.attempts} case={case} "
                          f"\nrowtest={rowtest} max={max} count={count} visible={visible} \nslice={sl} ==> {self.vector[slice(*sl)]}")
                                                  
            return not case
        
        self.vector[r*s + c] = move  # move probeweise eintragen
        res = all([
            check(self.skyline["west"][r],  [r*s, (r+1)*s, 1]),
            check(self.skyline["east"][r],  [-(s-r-1)*s-1, -(s-r)*s-1, -1]),
            check(self.skyline["north"][c], [c, c+s*s, s]),
            check(self.skyline["south"][c], [-(s-c), -((s+1)*s-c), -s])
        ])
        
        if self.test and (r+1==self.size or c+1==self.size) and res:
            self.show(f"valid={res}, level={level}, r={r}, c={c},  move={move}, attempt={self.attempts}")

        self.vector[r*s + c] = 0  # move wieder entfernen

        return res
            
    def record(self, level:int, move:int):
        self.vector[level]=move
        if not level in self.known:  # Vorgegebene Felder sind schon gesetzt
            r, c = self.rowAndCol(level)        
            self.rset[r].add(move)  # Wert zur Menge der Zeilenwerte hinzufügen
            self.cset[c].add(move)  # Wert zur Menge der Spaltenwerte hinzufügen
        

    def undo(self, level:int, move:int):
        self.vector[level]=0     
        if not level in self.known:  # Vorgegebene Felder nicht löschen
            r, c = self.rowAndCol(level)
            self.rset[r].remove(move)  # Wert aus der Menge der Zeilenwerte entfernen
            self.cset[c].remove(move)  # Wert aus der Menge der Spaltenwerte entfernen        

    def done(self, level: int):
        if level==self.size*self.size-1:
            self.resultatnumber += 1
            self.show(f"Resultat {self.resultatnumber:4d}, {self.attempts:18_d}. Versuch")
            return True
        return False

    def show(self, text):
        print(f"\n{text}")
        for r in range(self.size):
            for c in range(self.size):
                print(f"{self.vector[r*self.size + c]:4d}", end="")
            print()


if __name__ == "__main__":
    # Kurier-Freizeit, 20. August 2022
    skyline = {
        "west": [4, 0, 3, 2, 2],
        "east": [0, 3, 2, 0, 2],
        "north":[0, 4, 0, 0, 2],
        "south":[0, 0, 3, 3, 0]
    }
    known = {11:4, 18:2}
    Skyline(high=1, skyline=skyline, known=known, test=False, showAttempts=True)    


