# File sudoku.py
from backtrack import Backtrack
import math

class Sudoku(Backtrack):

    vector = []
    size = 9
    sqrtSize = int(math.sqrt(size))
    solutionNr = 0
    

    rset = [set() for i in range(size)]  # row sets
    cset = [set() for i in range(size)]  # column sets
    sset = [set() for i in range(size)]  # subsquare sets

    sudoku = { # leicht
        0:3, 4:1,
        12:5, 14:6, 15:9, 16:8,
        20:9, 25:1, 26:5,
        28:9, 29:4, 30:3, 31:6, 32:7, 34:5, 35:2,
        37:1, 38:7, 40:9, 41:5, 42:3, 43:4, 44:8,
        46:2, 47:3, 51:6,
        54:4, 56:6, 58:5, 59:2,
        63:9, 66:1, 68:3, 71:4,
        74:5, 78:8
        }

    values = range(1, size+1)

    def __init__(self, test:bool= False, showAttempts:bool = False):        
        self.vector = [0 for i in range(self.size*self.size)]
        super().__init__(test, showAttempts)

    def getSets(self, level:int):
        s = self.size
        sq = self.sqrtSize

        # Zeilennummer bestimmen: 0 bis size-1
        rr = level // s

        # Spaltennummer bestimmen: 0 bis size-1
        cc = level % s

        # Nummer des Unterquarats bestimmen: 0 bis size-1
        ss = rr // sq * sq + cc // sq

        return [self.rset[rr], self.cset[cc], self.sset[ss]]

    def moves(self, level: int):
        res = self.sudoku.get(level, 0)
        if res:
            return [res] 
        return list(set(self.values) - set().union(*self.getSets(level)))
    
    def valid(self, level: int, move) -> bool:
        return move not in set().union(*self.getSets(level))

    def done(self, level: int):
        if level==self.size*self.size-1:
            self.solutionNr += 1
            self.show(f"Resultat {self.solutionNr:4d}")
            return True
        return False        
    
    def record(self, level: int, move): 
        self.vector[level]=move  
        for s in self.getSets(level):
            s.add(move)        

        if self.test:
            print (level, move, rr, cc, ss, self.rset[rr], self.cset[cc], self.sset[ss])

    def undo(self, level: int, move):
        self.vector[level]=0
        for s in self.getSets(level):
            s.discard(move)        

    def show(self, text):
        print(f"\n{text}")
        for r in range(self.size):
            print(self.vector[r*self.size : (r+1)*self.size])       
   
if __name__ == "__main__":
    Sudoku(test=False, showAttempts=True) 
