# File zeltlager.py
from backtrack import Backtrack

class Zeltlager(Backtrack):
 
    trees = [0, 10, 14, 17, 23]
    rows =  [1, 1, 1, 1, 1]
    cols =  [1, 1, 2, 0, 1]
   
    # Auch rechteckige Plätze sind möglich, daher zwei Größenangaben
    rsize = len(rows)
    csize = len(cols)
    
    tents = []  # Zu Beginn sind keine Zelte aufgebaut
    solutionNr = 0      # Nummer der Lösung
    rowscount = [0 for i in range(rsize)]  # Aktuelle Anzahl der Zelte pro Zeile
    colscount = [0 for i in range(csize)]  # Aktuelle Anzahl der Zelte pro Spalte
 
    def __init__(self, test:bool= False, showAttempts:bool = False): 
        super().__init__(test, showAttempts)

    def rowAndCol(self, where:int) -> tuple:
        s = self.rsize
        return (where // s, where % s)

    def neighbors(self, place:int) -> list[int]:
       
        r, c = self.rowAndCol(place)        
        s = self.rsize
        n = [ ]    
        
        if c+1<s:
            n.append(place+1)            
        if r>0:
            n.append(place-s)        
        if c>0:
            n.append(place-1)
        if r+1<s:
            n.append(place+s)
        return n
    
    def moves(self, level: int) -> list[int]:        
        return self.neighbors(self.trees[level])

    def valid(self, level: int, move:int) -> bool:
        if move in self.trees:  # da steht ein anderer Baum
            return False
        
        if move in self.tents:  # da steht schon ein zelt
            return False

        for n in self.neighbors(move):
            if n in self.tents: # da steht schon ein Zelt daneben
                return False

        r, c = self.rowAndCol(move)
        if self.rowscount[r] >= self.rows[r]:   # kein weiteres Zelt möglich
            return False
        if self.colscount[c] >= self.cols[c]:   # kein weiteres Zelt möglich
            return False
        return True

    def record(self, level: int, move:int):
        self.tents.append(move)
        r, c = self.rowAndCol(move)
        self.rowscount[r] += 1
        self.colscount[c] += 1

    def undo(self, level: int, move:int):
        self.tents.pop()
        r, c = self.rowAndCol(move)
        self.rowscount[r] -= 1
        self.colscount[c] -= 1

    def done(self, level: int)-> bool:
        if len(self.trees) == len(self.tents):
            self.solutionNr += 1            
            self.show()
            return True
        return False

    def show(self):

        # Einfache Ausgabe:
        # print(f"\nSolution {self.solutionNr:4} level={level} {self.tents}")

        # Schönere Ausgabe:
        tree = "B"  #  B steht für Baum
        tent = "Z"  #  Z steht für Zelt
        print(f"\nLösung {self.solutionNr:4}")
        print(" "*6, end="")
        for c in range(self.csize):
            print(f"{self.cols[c]:3}", end="")
        print()
        for r in range(self.rsize):
            print(f"{self.rows[r]:2d} -> ", end="")
            for c in range(self.csize):
                f = r*self.rsize + c # Feldnummer
                if f in self.trees:
                    print(f"  {tree}", end="")
                elif f in self.tents:
                    print(f"  {tent}", end="")
                else:
                    print("  .", end="")
            print()

if __name__ == "__main__":
    Zeltlager(test=False, showAttempts=True)
