# File zeltlager_2.py
from backtrack import Backtrack

class Zeltlager(Backtrack):

    tents = []  # Zu Beginn sind keine Zelte aufgebaut
    treeNr = {} # Nummer des Baumes
    solutionNr = 0      # Nummer der Lösung

    def __init__(self, zeltlager={}, test:bool= False, showAttempts:bool = False):
        self.zeltlager = zeltlager
           
        # Auch rechteckige Plätze sind möglich, daher zwei Größenangaben
        self.rsize = len(zeltlager["rows"])
        self.csize = len(zeltlager["cols"])

        self.rowscount = [0 for i in range(self.rsize)]  # Aktuelle Anzahl der Zelte pro Zeile
        self.colscount = [0 for i in range(self.csize)]  # Aktuelle Anzahl der Zelte pro Spalte
 
    
        super().__init__(test, showAttempts)

    def rowAndCol(self, where):
        s = self.rsize
        return (where // s, where % s)

    def neighbors(self, place):
       
        r, c = self.rowAndCol(place)        
        s = self.rsize
        n = [ ]    
        
        if c+1<s:
            n.append(place+1)            
        if r>0:
            n.append(place-s)        
        if c>0:
            n.append(place-1)
        if r+1<s:
            n.append(place+s)
        return n
    
    def moves(self, level: int):        
        return self.neighbors(self.zeltlager["trees"][level])

    def valid(self, level: int, move) -> bool:
        if move in self.zeltlager["trees"]:  # da steht ein anderer Baum
            return False
        
        if move in self.tents:  # da steht schon ein zelt
            return False

        for n in self.neighbors(move):
            if n in self.tents: # da steht schon ein Zelt daneben
                return False

        r, c = self.rowAndCol(move)
        if self.rowscount[r] >= self.zeltlager["rows"][r]:   # kein weiteres Zelt möglich
            return False
        if self.colscount[c] >= self.zeltlager["cols"][c]:   # kein weiteres Zelt möglich
            return False
        return True

    def record(self, level: int, move):
        self.tents.append(move)
        self.treeNr[move] = level
        r, c = self.rowAndCol(move)
        self.rowscount[r] += 1
        self.colscount[c] += 1

    def undo(self, level: int, move):
        self.tents.pop()
        del self.treeNr[move]
        r, c = self.rowAndCol(move)
        self.rowscount[r] -= 1
        self.colscount[c] -= 1

    def done(self, level: int):
        if len(self.zeltlager["trees"]) == len(self.tents):
            self.solutionNr += 1            
            self.show()
            return True
        return False

    def show(self):

        # Einfache Ausgabe:
        # print(f"\nSolution {self.solutionNr:4} level={level} {self.tents}")

        # Schönere Ausgabe:
        tree = "B"  # B steht für Baum
        tent = "Z"  # Z steht für Zelt
        print(f"\nLösung {self.solutionNr:4}")
        print(" "*6, end="")
        for c in range(self.csize):
            print(f" {self.zeltlager['cols'][c]:3}", end="")
        print()
        
        for r in range(self.rsize):
            print(f"{self.zeltlager['rows'][r]:2d} -> ", end="")
            for c in range(self.csize):
                f = r*self.rsize + c # Feldnummer
                if f in self.zeltlager["trees"]:
                    
                    print(f"   {tree}", end="")
                    
                elif f in self.tents:
                    print(f" {self.treeNr[f]:3d}", end="")
                    # print(f"   {tent}", end="")
                else:
                    print("   .", end="")
            print()

if __name__ == "__main__":
    # https://www.janko.at/Raetsel/Zeltlager/050.a.htm
    zeltlager = {
        "trees": [ 8, 11, 13, 18, 24, 28, 29, 33, 40, 48, 58,
                   61, 64, 68, 74, 81, 86, 88, 91, 94, 108, 114,
                   128, 132, 135, 137, 141, 143],
        "rows":  [ 2, 3, 1, 3, 2, 2, 3, 2, 4, 0, 5, 1],
        "cols":  [ 2, 4, 2, 2, 2, 3, 1, 3, 0, 5, 0, 4]
    }
    Zeltlager(zeltlager=zeltlager, test=False, showAttempts=True)
